/********************************************************
 ** 2gif - SMS2/QDOS to GIF (c) Jonathan Hudson 1994
 **
 ** This program is derived from GIF code in Ghostscript
 ** The original copyright notices are reproduced below
 ********************************************************/


/* Copyright (C) 1992, 1993 Aladdin Enterprises.  All rights reserved.

This file is part of Ghostscript.

Ghostscript is distributed in the hope that it will be useful, but
WITHOUT ANY WARRANTY.  No author or distributor accepts responsibility
to anyone for the consequences of using it or for whether it serves any
particular purpose or works at all, unless he says so in writing.  Refer
to the Ghostscript General Public License for full details.

Everyone is granted permission to copy, modify and redistribute
Ghostscript, but only under the conditions described in the Ghostscript
General Public License.  A copy of this license is supposed to have been
given to you along with Ghostscript so you can know your rights and
responsibilities.  It should be in a file named COPYING.  Among other
things, the copyright notice and this notice must be preserved on all
copies.  */

/* Thanks to Phil Conrad for donating the original version */
/* of these drivers to Aladdin Enterprises. */

/*======================================================================*/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <fcntl.h>
#include <unistd.h>
#include <sys/stat.h>


#ifdef __i386__
#define LITTLE_ENDIAN 1
#endif

#define swapword(v) ((ushort)(v) >> 8) + ((ushort)(v) << 8)

#ifdef LITTLE_ENDIAN
#define assign_ushort(v) v
#else
#define assign_ushort(v) swapword(v)
#endif

#define ASPECT87    1
#define ASPECT89    2
#define XVPATCH     4
#define ASPECTNONE  8

#define ASPECTMASK  0x1F

#define private static

typedef unsigned char byte;
#ifndef __unix__
typedef unsigned short ushort;
typedef unsigned int uint;
typedef unsigned long ulong;
#endif

typedef struct
{
    ushort flag;
    short xpix;
    short ypix;
    short llen;
    short zero;
}
QHEAD, *PQHEAD;

typedef struct
{
    int fd;
    FILE *fp;
    QHEAD qh;
    short flag;
    short qlar;
}
QLS, *PQLS;

#define globalcolor_shift 7
#define colorres_shift 4
#define sort_shift 3
#define colorsize_shift 0

typedef struct gif_header_s
{
    byte signature[3];		/* magic number == 'GIF' */
    byte version[3];		/* version # '87a' or '89a' */
    ushort width;		/* screen width */
    ushort height;		/* screen height */
#if 0
      struct  {               /* bit structure of flags */
      unsigned globalcolor:1; /* global color table flag - MSB */
      unsigned colorres:3;    /* bits/color */
      unsigned sort:1;                /* color table sorted */
      unsigned colorsize:3;   /* 2^colorsize bytes in color table -LSB */
      } flags;
#endif

    byte flags;
    byte background;		/* background color index */
    byte aspect;		/* pixel aspect ratio */
/* ratio = (aspect + 15) / 64 */
}
gif_header;

typedef struct image_descriptor_s
{
#if 0
    byte    separator;      /* image separator == 0x2c */
#endif
    ushort left_pos;		/* image left pos (pixels) */
    ushort top_pos;		/* image top  pos (pixels) */
    ushort width;		/* image width    (pixels) */
    ushort height;		/* image height   (pixels) */
#if 0
      struct  {               
      unsigned localcolor:1;  /* local color table flag */
      unsigned interlace:1;   /* image interlaced  0=no */
      unsigned sort:1;        /* color table sorted 0=no */
      unsigned resv:2;        */
      unsigned localsize:3;   /* 2^localsize+1 = color table size */
      } flags;
#endif
    byte flags;
}
image_descriptor;

/********************************************************/
/* LZW routines are based on:                           */
/* Dr. Dobbs Journal --- Oct. 1989.                     */
/* Article on LZW Data Compression by Mark R. Nelson    */
/********************************************************/

#define MAX_BITS 12		/* this is max for GIF. */

#define TABLE_SIZE 5123		/* this is max for 12-bit codes */
#define TABLE_HASH_SHIFT 2	/* size < 4095 + (4095 >> shift) */

/* State of LZW encoder */
typedef struct code_entry_s
{
    int code_value;
    ushort prefix_code;
    byte append_character;
}
code_entry;
typedef struct lzw_encoder_s
{
    int bits;
    ushort Max_Code;
    ushort Clear_code;
    ushort next_code;
    FILE *file;
    code_entry *table;
    ushort string_code;
/* State of output buffer */
    byte output_bit_buffer;
    int output_bit_count;	/* # of valid low-order bits */
/* (between 0 and 7) in buffer */
    uint byte_count;
    byte gif_buffer[260];
}
lzw_encoder;

/* Initialize LZW encoder */
private void lzw_set_bits (register lzw_encoder *, int);
private void lzw_reset (register lzw_encoder *);
private int
  lzw_init (register lzw_encoder * pe, int bits, FILE * file)
{
    lzw_set_bits (pe, bits);
    pe->Clear_code = (1 << bits);
    pe->file = file;
    pe->byte_count = 1;
    pe->output_bit_count = 0;
    pe->output_bit_buffer = 0;
    pe->table = (code_entry *) malloc (TABLE_SIZE * sizeof (code_entry));

    if (pe->table == 0)
	return -2;

    lzw_reset (pe);
    pe->string_code = 0;
    return 0;
}
/* Establish the width of the code in bits */
private void
  lzw_set_bits (register lzw_encoder * pe, int bits)
{
    pe->bits = bits;
    pe->Max_Code = (1 << (bits + 1)) - 1;
}
/* Reset the encoding table */
private void
  lzw_reset (register lzw_encoder * pe)
{
    int index;

    for (index = 0; index < TABLE_SIZE; index++)
	pe->table[index].code_value = -1;
    pe->next_code = pe->Clear_code + 2;
}

/* Put out (data) of length (bits) to GIF buffer */
private void
  lzw_putc (register lzw_encoder * pe, uint data)
{
    int bits = pe->bits + 1;	/* output width */
    ulong buffer = pe->output_bit_buffer | ((ulong) data << pe->output_bit_count);

    pe->output_bit_count += bits;
    while (pe->output_bit_count >= 8)
    {				/* putc(output_bit_buffer >> 24, file); */
	pe->gif_buffer[pe->byte_count] = (byte) buffer;		/* low byte */
	buffer >>= 8;
	pe->output_bit_count -= 8;
	pe->byte_count++;
	if (pe->byte_count == 256)
	{
	    pe->byte_count = 1;
	    pe->gif_buffer[0] = 255;	/* byte count for block */
	    fwrite (pe->gif_buffer, 1, 256, pe->file);
	}
    }
    pe->output_bit_buffer = (byte) buffer;
}

/* Finish encoding, and flush the buffers. */
private void
  lzw_finish (register lzw_encoder * pe)
{
    lzw_putc (pe, pe->string_code);	/* output last code */
    lzw_putc (pe, pe->Clear_code + 1);	/* output eof code */
    lzw_putc (pe, 0);		/* force out last code */
    if (pe->byte_count != 1)
    {
	pe->gif_buffer[0] = pe->byte_count;
	fwrite (pe->gif_buffer, 1, pe->byte_count + 1, pe->file);
    }
}

/* Terminate LZW encoder. */
private void
  lzw_exit (register lzw_encoder * pe)
{
    free ((char *) pe->table);
}

/* Get the next (depth) bits from the pixel buffer. */
/* Note that 8 % depth == 0. */
/* Free variables: bits_left, bit_buffer, next, depth, depth_mask. */
#define lzw_getc()\
 (bits_left =\
  (bits_left == 0 ?\
   (bit_buffer = *(next++), 8 - (depth)) :\
   bits_left - (depth)),\
  (bit_buffer >> bits_left) & (depth_mask))

/* Output 1 row of data in GIF (LZW) format. */
private void
  lzw (byte * from, byte * end, register lzw_encoder * pe, int depth)
{
    int bits_left = 0;
    uint bit_buffer = 0;
    byte *next = from;
    uint depth_mask = (1 << depth) - 1;

    if (pe->next_code == (pe->Clear_code + 2))	/* first time through */
    {
	pe->string_code = lzw_getc ();
    }

    while (next < end || bits_left >= depth)
    {
	uint data = lzw_getc ();	/* actually only a byte */

    /* Hash to find a match for the prefix+char */
    /* string in the string table */

	ushort hash_prefix = pe->string_code;
	int index = (data << 4) ^ hash_prefix;
	int hash_offset;
	register code_entry *pce;

	index += index >> TABLE_HASH_SHIFT;
	if (index == 0)
	    hash_offset = 1;
	else
	    hash_offset = TABLE_SIZE - index;

	while (1)
	{
	    pce = &pe->table[index];
	    if (pce->code_value == -1)
		break;
	    if (pce->prefix_code == hash_prefix &&
		pce->append_character == data)
		break;
	    index -= hash_offset;
	    if (index < 0)
		index += TABLE_SIZE;
	}
	if (pce->code_value != -1)
	    pe->string_code = pce->code_value;
	else
	{			/* Make a new entry */
	    pce->code_value = pe->next_code++;
	    pce->prefix_code = pe->string_code;
	    pce->append_character = data;

	    lzw_putc (pe, pe->string_code);

	    if (pe->next_code > (pe->Max_Code + 1))
	    {			/* Increment the width of the code */
		if (pe->bits + 1 >= MAX_BITS)
		{		/* output clear code first */
		    lzw_putc (pe, pe->Clear_code);
		    pe->bits = (depth == 1 ? 2 : depth);
		    lzw_reset (pe);
		}
		else
		    pe->bits++;
		lzw_set_bits (pe, pe->bits);
	    }
	    pe->string_code = data;
	}
    }
}

static void MakeChunkyPix (byte * cpix, byte * qpix, short xpix)
{
    byte *q, *p;
    ushort k;
    ushort j;
    ushort l;
    ushort gb, rb;
    short llen, ns, m;
	
    llen = ((xpix+3) & ~3) >> 2;

    ns = (xpix & 3);
    
    p = cpix;
    q = qpix;
    memset (cpix, '\0', llen);
    l = k = 7;

    for (m = 0; m < xpix; m++)
    {
	j = 1 << k;
	gb = (*q & j);
	if (gb)
	{
	    *p |= (1 << l);
	}

	l--;

	rb = *(q + 1) & j;
	if (rb)
	{
	    *p |= (1 << l);
	}

	if (l == 0)
	{
	    p++;
	    l = 7;
	}
	else
	{
	    l--;
	}

	if (k == 0)
	{
	    k = 7;
	    q += 2;
	}
	else
	{
	    k--;
	}
    }
/*************
    if(ns)
    {
        byte a,b;
        static byte xx[] = {0xc0, 0x30, 0xc, 0x3};
        a = *p;
        
        for(i = ns; i < 4; i++)
        {
            if((b = (a & xx[ns-1])))
            {
                a |= b >> 2;
            }
        }
        *p = a;
    }
 ************/
}


/* Write a page to a file in GIF format. */

int DoGif (PQLS pq)
{
    static unsigned char qlcols[12] =
    {0,0,0, 0xff,0,0, 0,0xff,0, 0xff,0xff,0xff};
    static byte txt[] = "!\xfe\0352gif (c) Jonathan Hudson 1994";
    short height;
    short width;

    byte *qldat = (byte *) malloc (pq->qh.llen);
    byte *row = (byte *) malloc (pq->qh.llen * 2);
    byte *end = row + pq->qh.llen;

    gif_header header;
    image_descriptor header_desc;
    lzw_encoder encoder;
    int y;
    int k;
    int code = 0;		/* return code */

    if (row == 0)		/* can't allocate row buffer */
	return -2;

    code = lzw_init (&encoder, 2, pq->fp);

    if (code < 0)
	return code;

    /* Set up the header. */

    memcpy (header.signature, "GIF", 3);
    memcpy (header.version, "87a", 3);
    width = (pq->qh.xpix+3) & ~3;
    header.width = assign_ushort (width);
    height = pq->qh.ypix;
    if (pq->flag == 0)
    {
	height = height * pq->qlar / (pq->qlar - 1);
    }
    header.height = assign_ushort (height);

    header.flags =
	(1 << globalcolor_shift) +
	(1 << colorres_shift) +
	(0 << sort_shift) +
	(1 << colorsize_shift);
    header.background = 0;
    if (pq->flag & ASPECT89)
    {
	header.aspect = ((64 * (pq->qlar - 1)) / pq->qlar) -15;
	header.version[1] = '9';
    }
    else
    {
	header.aspect = 0;
    }

    /* Write the header, on the first page only. */

    if (fwrite (&header, 1, 13, pq->fp) < 13)
    {
	code = -2;
	goto gif_done;
    }

    /* Write the header global color palette. */
    fwrite (qlcols, 1, 12, pq->fp);

    if (pq->flag & ASPECT87)
    {
	static char aext[6] =
	{'!', 'R', 2, 0, 0, 0};
        aext[3] = pq->qlar-1;
        aext[4] = pq->qlar;
	fwrite (aext, 1, (pq->flag & XVPATCH) ? 5 : 6, pq->fp);
    }

    fwrite (txt, 1, 33, pq->fp);

    header_desc.left_pos = 0;
    header_desc.top_pos = 0;
    header_desc.width = header.width;
    header_desc.height = header.height;

    header_desc.flags = 1;

    fputc (0x2c, pq->fp);	/* start with separator */
    if (fwrite (&header_desc, 1, 9, pq->fp) < 9)
    {
	code = -2;
	goto gif_done;
    }

    fputc (encoder.bits, pq->fp);	/* start with code size */

    lzw_putc (&encoder, encoder.Clear_code);	/* output clear code first */

    /* Dump the contents of the image. */

    end = row + (width >> 2);

    for (k = y = 0; k < height;)
    {
        k++;
        if(pq->flag || ((k % pq->qlar) && y < pq->qh.ypix))
        {
            read (pq->fd, qldat, pq->qh.llen);
            y++;
        }
	MakeChunkyPix (row, qldat, pq->qh.xpix);
	lzw (row, end, &encoder, 2);
    }
    lzw_finish (&encoder);
    fputc (0, pq->fp);

  gif_done:
    lzw_exit (&encoder);
    free ((char *) row);
    free ((char *) qldat);

    if (pq->fp != NULL)
    {
	fputc (';', pq->fp);	/* EOF indicator */
    }

    return code;
}

void ObnoxiousMsg (void)
{
    fprintf (stderr, "The Graphics Interchange Format(c) is\n");
    fprintf (stderr, "the Copyright Property of CompuServe Incorporated.\n");
    fprintf (stderr, "GIF(sm) is a Service Mark property of CompuServe Incorporated.\n");
}

void usage (void)
{
    fputs (" 2gif  [-7] [-9] [-i] [-a] in out\n", stderr);
    fputs ("        -7 generate GIF87 a/r\n", stderr);
    fputs ("        -9 generate GIF89 a/r\n", stderr);
    fputs ("        -a 'nicer' a/r\n", stderr);
    fputs ("        -i no a/r\n", stderr);
    fputs ("           stdin & stdout if no files given\n", stderr);
    exit (0);
}

int main (int ac, char **av)
{
    QLS qs;
    struct stat s;

    char *ip = NULL;
    char *op = NULL;
    
    qs.qlar = 4;
    qs.flag = 0;

    if(ac == 1)
    {
	usage();
    }
    
    
    while(++av,--ac)
    {
        if(**av == '-')
        {
            switch(*++(*av))
            {
	      case 'a':
		qs.qlar = 3;
		break;
	      case 'i':
		qs.flag = ASPECTNONE;
		break;
	      case '7':
		qs.flag = ASPECT87;
		if(*++(*av) == 'x')
		{
		    qs.flag |= XVPATCH;
		}
		break;
	      case '9':
		qs.flag = ASPECT89;
		break;
	      case 'h':
	      case '?':
		usage();
		break;
	    }
	}
	else
	{
            if(ip == NULL)
            {
                ip = *av;
            }
            else if(op == NULL)
            {
                op = *av;
            }
        }
    }

    
    if (ip)
    {
	qs.fd = open (ip, O_RDONLY);
    }
    else
    {
	qs.fd = fileno(stdin);
    }
    if(qs.fd > 0)    
    {
	read (qs.fd, &qs.qh, sizeof (QHEAD));
#ifdef LITTLE_ENDIAN
        qs.qh.flag = swapword(qs.qh.flag); 
        qs.qh.xpix = swapword(qs.qh.xpix); 
        qs.qh.ypix = swapword(qs.qh.ypix); 
        qs.qh.llen = swapword(qs.qh.llen); 
#endif

	if (qs.qh.flag != 0x4afc)
	{
	    fprintf(stderr, "Screen !! %x", qs.qh.flag);
	    fstat (qs.fd, &s);
	    qs.qh.xpix = 512;
	    qs.qh.llen = 128;
	    qs.qh.ypix = (s.st_size + 127) / 128;
	    lseek (qs.fd, 0, 0);
	}

	if(op)
	{
	    qs.fp = fopen (op, "w");
	}
	else
	{
	    qs.fp = stdout;
	}
	
	if (qs.fp)
	{
	    ObnoxiousMsg ();
	    fprintf(stderr,"Size %d x %d\n", qs.qh.xpix, qs.qh.ypix);
	    DoGif (&qs);
	    fclose (qs.fp);
	}
	else
	{
	    perror ("output");
	}
    }
    else
    {
	perror ("input");
    }
    return 0;
}

